/* ==========================================================================*/
/* - - - - - - - - -  W s p r e n c o d e r. c  - - - - - - - - - - - - - -  */
/* ==========================================================================*/
/*    Implementation File for the wsprencoder module                         */
/*	Created   2011-09-10   M. Wheatley                                       */
/*	Modified  2011-09-10   MW                                                */
/*     wispr data encoder                                                    */
///////////////////////////////////////////////////////////////////////////////
//            2011 Copyright (C) Moe Wheatley, AE4JY
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////////////

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include "wsprencoder.h"
#include <string.h>

/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/

//WSPR character coding
#define W_NDIGITS 10		//10 digits 0 to 9
#define W_NLETTERS 26		//26 letters A to Z  Caps only
#define W_NLETTERSPACE 27	//26 letters plus space
#define W_NGRIDLETTERS 18	//18 letters A to R

// Convolutional coding polynomials. All are rate 1/2, K=32
// Layland-Lushbaugh code
// Nonsystematic, non-quick look-in, dmin=?, dfree=?
#define	POLY1	0xF2D05351
#define	POLY2	0xE4613C47


/*---------------------------------------------------------------------------*/
/*-------------------> M E M O R Y  A L L O C A T I O N <--------------------*/
/*---------------------------------------------------------------------------*/
static uint8_t SrcBytes[NUM_SRCBYTES];
static uint8_t TmpSymbols[NUM_SRCBYTES*8*2];
uint8_t Symbols[NUM_SYMBOLS];

/* 8-bit parity lookup table (XOR of all the bits) */
const uint8_t PARTBL[256] = {
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
};

const uint8_t INTERLEAVE_TBL[NUM_SYMBOLS] = {
  0 ,128 , 64 , 32 ,160 , 96 , 16 ,144 , 80 , 48 ,112 ,  8 ,136 , 72 , 40 ,104 , 24 ,152 , 88 , 56 ,
120 ,  4 ,132 , 68 , 36 ,100 , 20 ,148 , 84 , 52 ,116 , 12 ,140 , 76 , 44 ,108 , 28 ,156 , 92 , 60 ,
124 ,  2 ,130 , 66 , 34 , 98 , 18 ,146 , 82 , 50 ,114 , 10 ,138 , 74 , 42 ,106 , 26 ,154 , 90 , 58 ,
122 ,  6 ,134 , 70 , 38 ,102 , 22 ,150 , 86 , 54 ,118 , 14 ,142 , 78 , 46 ,110 , 30 ,158 , 94 , 62 ,
126 ,  1 ,129 , 65 , 33 ,161 , 97 , 17 ,145 , 81 , 49 ,113 ,  9 ,137 , 73 , 41 ,105 , 25 ,153 , 89 ,
 57 ,121 ,  5 ,133 , 69 , 37 ,101 , 21 ,149 , 85 , 53 ,117 , 13 ,141 , 77 , 45 ,109 , 29 ,157 , 93 ,
 61 ,125 ,  3 ,131 , 67 , 35 , 99 , 19 ,147 , 83 , 51 ,115 , 11 ,139 , 75 , 43 ,107 , 27 ,155 , 91 ,
 59 ,123 ,  7 ,135 , 71 , 39 ,103 , 23 ,151 , 87 , 55 ,119 , 15 ,143 , 79 , 47 ,111 , 31 ,159 , 95 ,
 63 ,127
};

const uint8_t SYNC_TBL[NUM_SYMBOLS] = {
1,1,0,0,0,0,0,0,1,0,0,0,1,1,1,0,0,0,1,0,0,1,0,1,1,1,1,0,0,0,0,0,0,0,1,0,0,1,0,1,0,0,
0,0,0,0,1,0,1,1,0,0,1,1,0,1,0,0,0,1,1,0,1,0,0,0,0,1,1,0,1,0,1,0,1,0,1,0,0,1,0,0,1,0,
1,1,0,0,0,1,1,0,1,0,1,0,0,0,1,0,0,0,0,0,1,0,0,1,0,0,1,1,1,0,1,1,0,0,1,1,0,1,0,0,0,1,
1,1,0,0,0,0,0,1,0,1,0,0,1,1,0,0,0,0,0,0,0,1,1,0,1,0,1,1,0,0,0,1,1,0,0,0
};

static void ConvolutionalEncode();
static void Interleave();
static void AddSyncVector();

///////////////////////////////////////////////////////////////////////////////
//   Encode up to 6 character CallSign, 4 character GridSquare, and dBm Power
// into a WSPR Symbol array
///////////////////////////////////////////////////////////////////////////////
int WsprEncode(char* CallSign, char* GridSquare, int dBm )
{
int i,j;
uint32_t 	m;
uint32_t 	n;
char buf[6+1];
char* p = buf;	//create local pointer to buffer
//printf("\r\n%s   %s  %i  \r\r", CallSign, GridSquare, dBm);
	//sanity check on callsign length
	n = strlen(CallSign);
	if( (n<3) || (n>16) )
		return 0;	//error if is callsign string too short
	//fill buffer with spaces
	for(i=0; i<6; i++)
		buf[i] = ' ';
	//first copy callsign into buf[] and remove any spaces on either side
	j = 0;
	for(i=0; i<n; i++)
	{
		if(CallSign[i] != ' ')
			buf[j++] = CallSign[i];
		if(j>6)
			return 0;	//string too long
	}
	buf[6] = 0;		//zero terminate at 6 chars
	//buf[] now is a 6 character string left justified and padded with spaces on the end if needed
//printf("\r\nCall .%s.\r\n", p);
	//process first character in callsign
	//first check if 3rd character is a number else just use space value for first position
	if( (p[2]>='0') && (p[2]<='9') )	//if 3rd character not a digit then prefix with space
	{
		if( (*p>='0') && (*p<='9') )	//if first character is a digit
			n = (*p - '0');
		else if((*p>='A') && (*p<='Z') )//if a letter
			n = (*p - 'A' + W_NDIGITS);
		else if(*p==' ')
			n = W_NDIGITS + W_NLETTERS;	//space code for first character
		else
			return 0;	//is not valid character
		p++;	//point to next char in callsign
	}
	else
	{	//if 2nd character is a number then pad in a space at the beginning
		if( (p[1]>='0') && (p[1]<='9') )
			n = W_NLETTERS+W_NDIGITS;
		else
			return 0;	//is not valid character
	}
//printf("N1 %d\r\n", n);
	//process second character in callsign
	n *= (W_NLETTERS+W_NDIGITS);
	if( (*p>='0') && (*p<='9') )	//if first character is a digit
		n += (*p - '0');
	else if((*p>='A') && (*p<='Z') )
		n += (*p - 'A' + W_NDIGITS);
	else
		return 0;	//is space or not valid character
	p++;	//point to next char in callsign
//printf("N2 %d\r\n", n);

	//process third character in callsign which must be numeric
	n *= W_NDIGITS;
	if( (*p>='0') && (*p<='9') )	//if first character is a digit
		n += (*p - '0');
	else
		return 0;	//is space or not valid character
	p++;	//point to next char in callsign
//printf("N3 %d\r\n", n);

	//process last 3 characters 4 thru 6 in callsign which must be letter or space
	for(i=0; i<3; i++)
	{
		n *= W_NLETTERSPACE;
		if((*p>='A') && (*p<='Z') )
			n += (*p - 'A');
		else if(*p==' ')
			n += W_NLETTERS;		//space code for letter only field
		else
			return 0;	//is not valid
		p++;	//point to next char in callsign
//printf("N456 %d\r\n", n);
	}
	//
	// n should now be the 28 packed callsign bits
	//
	//Now encode the 4 character gridsquare
	//first sanity check the gridsquare for 2 letters and 2 digits
	p = GridSquare;
	if( (p[0]<'A') || (p[0]>'R') || (p[1]<'A') || (p[1]>'R') )
		return 0;
	if( (p[2]<'0') || (p[2]>'9') || (p[3]<'0') || (p[3]>'9') )
		return 0;
	m = 180*(179 - 10*(p[0] - 'A') - (p[2] - '0') ) + 10*(p[1] - 'A') + (p[3] - '0');
	//now add the dBm power value
	m = m*128 + dBm + 64;
	//concatenate 28 bits of n with 22 bits of m to get left justified 50 bits in SrcWords[]
	for(i=0; i<11; i++)
		SrcBytes[i] = 0;
	//left justify the 28 bits of n into SrcBytes[] array
	SrcBytes[0] = (n>>20);			//bits 27-20 goto 7-0
	SrcBytes[1] = (n>>12)&0x00FF;	//bits 19-12 goto 7-0
	SrcBytes[2] = (n>>4)&0x00FF;	//bits 11-4 goto 7-0
	SrcBytes[3] = (n<<4)&0x00FF;	//bits 3-0 goto 7-4

	//left justify the 22 bits of m into SrcBytes[] array
	SrcBytes[3] |= ((m>>18)&0x0FF);	//OR in top 4 bits of m into last 4 bits of SrcBytes[3] bits 21-18 goto 3-0
	SrcBytes[4] = (m>>10);			//bits 17-10 goto 7-0
	SrcBytes[5] = (m>>2);			//bits 9-2 goto 7-0
	SrcBytes[6] = (m<<6)&0xFF;		//bits 1-0 goto 7-6
	//
	//here with SrcBytes[] containing left(msb) justified 50 bit input data with 0's padded to end
//printf("\r\n");
//	for(i=0; i<7; i++)
//		printf("%X ", SrcBytes[i]);
//printf("\r\n");
	ConvolutionalEncode();
	Interleave();
	AddSyncVector();
	return 1;
}


///////////////////////////////////////////////////////////////////////////////
// Convolutionally encode a packet. The input data bytes are read
// high bit first and the encoded packet is written into 'symbols',
// one symbol per byte. The first symbol is generated from POLY1,
// the second from POLY2.
//
// Storing only one symbol per byte uses more space, but it is faster
// and easier than trying to pack them more efficiently.
///////////////////////////////////////////////////////////////////////////////
static void ConvolutionalEncode()
{
uint32_t encstate;
uint32_t tmp;
int i,j,k;
	encstate = 0;
	k = 0;
	for(j=0; j<NUM_SRCBYTES; j++)
	{
		for(i=7; i>=0; i--)
		{	//create convolution symbols from input data
			encstate = (encstate << 1) | ((SrcBytes[j] >> i) & 1);	//get ith bit
			tmp = encstate & POLY1;
			//calc 32 bit parity for tmp
			tmp ^= (tmp >> 16);
			TmpSymbols[k++] = PARTBL[(tmp ^ (tmp >> 8)) & 0xff];
			tmp = encstate & POLY2;
			tmp ^= (tmp >> 16);
			TmpSymbols[k++] = PARTBL[(tmp ^ (tmp >> 8)) & 0xff];
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Create interleaved symbols using table lookup as index into symbol array
///////////////////////////////////////////////////////////////////////////////
static void Interleave()
{
int i;
	for(i=0; i<NUM_SYMBOLS; i++)
		Symbols[INTERLEAVE_TBL[i]] = TmpSymbols[i];
}

///////////////////////////////////////////////////////////////////////////////
// Add the sync vector to the interleaved symbols
///////////////////////////////////////////////////////////////////////////////
static void AddSyncVector()
{
int i;
	for(i=0; i<NUM_SYMBOLS; i++)
		Symbols[i] = SYNC_TBL[i] + (Symbols[i]<<1);
#if 0
	for(i=0; i<NUM_SYMBOLS; i++)
	{
		printf("%i",Symbols[i]);
	}
	printf("\r\n");
#endif
}
